package de.dassit;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

import org.apache.commons.lang.StringUtils;

public class PreferencesTool implements Serializable {
	private static final long serialVersionUID = -7693367015487808983L;
	protected Properties preferences;
	protected String directory;
	protected String user;
	protected File prefsFile;

	public void initialize() throws IOException {
		if (directory == null)
			throw new IOException("Directory name not set");
		if (user == null)
			throw new IOException("User name not set");

		prefsFile = new File(directory, user + ".properties");
		preferences = new Properties();
		if (prefsFile.exists())
			preferences.load(new FileReader(prefsFile));
	}

	public void load() throws IOException {
		if (preferences != null) {
			return;
		}

		synchronized (this) {
			initialize();
		}
	}

	public String getUserName() {
		return user;
	}

	public void setUserName(String userName) {
		this.user = userName;
	}

	public String getDirectory() {
		return directory;
	}

	public void setDirectory(String directory) {
		this.directory = directory;
	}

	public String get(String key) throws IOException {
		String result = null;

		try {
			load();
			result = preferences.getProperty(key);
		} catch (IOException ex) {
			System.out
					.println("UserPrefsFileTool Error: IOException in load() for "
							+ makeKey(key) + " " + ex.getMessage());
		} catch (java.lang.InternalError err) {
			System.out
					.println("UserPrefsFileTool Error: InternalErrror during deserialization of "
							+ makeKey(key) + " " + err.getMessage());
		}
		return result;
	}

	public List<String> getMulti(String key) throws IOException {
		String v = get(key);
		if (v == null)
			return Collections.<String> emptyList();
		return Arrays.asList(v.split(","));
	}

	public boolean multiContains(String key, String value) throws IOException {
		List<String> values = getMulti(key);
		return values.contains(value);
	}

	public void put(String key, String value) throws IOException {
		load();
                String[] lines = value.split("[\r\n]");
                String trimmedValue = "";
                if(lines.length > 0)
                    trimmedValue = lines[0].trim();
		preferences.setProperty(key, trimmedValue);
		preferences.store(new FileWriter(prefsFile), "User Preferences");
	}

	public void putMulti(String key, List<String> values) throws IOException {
		put(key, StringUtils.join(values, ","));
	}
	
	public void putMulti(String key, String[] values) throws IOException {
		put(key, StringUtils.join(values, ","));
	}

	public String dump() throws IOException {
		StringBuilder out = new StringBuilder();
		load();
		String[] keys = preferences.keySet().toArray(new String[0]);
		Arrays.sort(keys);
		for (String key : keys) {
			out.append(key);
			out.append(":");
			out.append(preferences.get(key));
			out.append('\n');
		}
		return out.toString();
	}

	protected String makeKey(String key) {
		if (user == null)
			return key;
		return user + "." + key;
	}

	public static void main(String[] args) throws IOException {
		PreferencesTool tool = new PreferencesTool();
		tool.setDirectory("/tmp");
		tool.setUserName("testuser");
		tool.put("mypref", "test");
		String[] values = { "val1", "val2" };
		tool.putMulti("multi", Arrays.asList(values));
		System.out.println(tool.get("mypref"));
		System.out.println("dump:");
		System.out.println(tool.dump());
		System.out.println(tool.multiContains("multi", "val1"));
		System.out.println(tool.multiContains("multi", "val3"));
		System.out.println(tool.getMulti("multi"));
	}

}
