package de.dassit.dassadmin;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.logging.Logger;

import javax.naming.NamingException;
import javax.naming.directory.SearchResult;

import de.dassit.ldap.Connection;
import de.dassit.ldap.Entry;
import de.dassit.ldap.EntryList;

public class BadMailGroup extends Entry {
	private static final Logger log = Logger.getLogger(BadMailGroup.class
			.getName());
	
	private static String GROUP_MEMBER_MAIL_ADDRESS = "snitGroupMemberMailAddress";
	private static String SENDER_RESTRICTED_MAIL_ADDRESS = "snitSenderRestrictedMailAddress";
	private static String MEMBERSHIP = "snitDynamicGroupMember";

	static String userContainer = "ou=Users";

	public BadMailGroup() {
		setNamingParameters();
	}

	public BadMailGroup(SearchResult sr) {
		super(sr);
	}

	@Override
	protected void setNamingParameters() {
		setNamingAttribute(defaultNamingAttribute);
		setObjectClasses(new String[] { "domainRelatedObject", "snitMailGroup",
				"groupOfNames", "labeledURIObject" });
		setContainer(defaultContainer);
	}

	public boolean hasRestrictedSender(String s) {
		List<String> restrictedSenders = getRestrictedSenders();
		if (s == null || restrictedSenders == null)
			return false;
		for (String n : restrictedSenders) {
			if (n.equals(s))
				return true;
		}
		return false;
	}

	public List<String> getDestinations() {
		List<String> result = getAttributeMultiValue(GROUP_MEMBER_MAIL_ADDRESS);
		if (result == null)
			return null;
		Collections.sort(result);
		return result;
	}

	public void addDestination(String s) {
		appendAttributeValue(GROUP_MEMBER_MAIL_ADDRESS, s);
	}

	public void removeDestination(String s) {
		removeAttributeValue(GROUP_MEMBER_MAIL_ADDRESS, s);
	}

	public List<String> getRestrictedSenders() {
		List<String> result = getAttributeMultiValue(SENDER_RESTRICTED_MAIL_ADDRESS);
		if (result == null)
			return null;
		Collections.sort(result);
		return result;
	}

	public void addRestrictedSender(String s) {
		appendAttributeValue(SENDER_RESTRICTED_MAIL_ADDRESS, s);
	}

	public void removeRestrictedSender(String s) {
		removeAttributeValue(SENDER_RESTRICTED_MAIL_ADDRESS, s);
	}

	public void removeFromUsers() throws NamingException {
		List<DassUser> dynMembers = DassUser
				.searchDassUsers(MEMBERSHIP + "=" + this.getName(),
						getDefaultContainer(DassUser.class.getName()));

		for (DassUser u : dynMembers) {
			u.removeAttributeValue(MEMBERSHIP, this.getName());
			u.save();
		}
	}

	public void updateMembers(String oldName) throws NamingException {
		List<DassUser> dynMembers = DassUser.searchDassUsers(MEMBERSHIP + "="
				+ oldName, userContainer);
		log.info("updateMembers:" + dynMembers.size());
		for (DassUser u : dynMembers) {
			log.info(u.getName());
			u.removeAttributeValue(MEMBERSHIP, oldName);
		}
		for (DassUser u : dynMembers) {
			u.appendAttributeValue(MEMBERSHIP, this.getName());
			u.save();
		}
	}

	@Override
	public void delete() throws NamingException {
		removeFromUsers();
		super.delete();
	}

	@Override
	public void rename(String newName) throws NamingException {
		String oldName = getName();

		rename(Connection.getDefaultConnection(), newName);
		this.setName(newName);
		updateMembers(oldName);
	}

	public List<BadMailGroup> getAll() {
		return search("objectClass=snitMailGroup", defaultContainer);
	}

	public List<BadMailGroup> search(String filter, String container) {

		EntryList<BadMailGroup> entryList = new EntryList<BadMailGroup>() {
			public void addInstance(SearchResult sr) {
				entries.add(new BadMailGroup(sr));
			}
		};

		Connection.getDefaultConnection().search(filter, container, entryList);
		List<BadMailGroup> result = entryList.getEntries();

		Collections.sort(result, new Comparator<BadMailGroup>() {

			public int compare(BadMailGroup arg0, BadMailGroup arg1) {
				return arg0.getName().compareToIgnoreCase(arg1.getName());
			}
		});

		return entryList.getEntries();
	}

	public static String getGroupMemberMailAddress() {
		return GROUP_MEMBER_MAIL_ADDRESS;
	}

	public static String getSenderRestrictedMailAddress() {
		return SENDER_RESTRICTED_MAIL_ADDRESS;
	}

	public static String getMembership() {
		return MEMBERSHIP;
	}

	@Override
	protected EntryList<? extends Entry> getEntryList() {
		return new EntryList<BadMailGroup>() {
			public void addInstance(SearchResult sr) {
				entries.add(new BadMailGroup(sr));
			}
		};
	}
}
