package de.dassit.dassadmin;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import javax.naming.NamingException;

public class DeviceType {
	private static final Logger log = Logger.getLogger(DeviceType.class.getName());

	public static final String CONFIGFILE_SUFFIX = ".properties";
	protected String name;
	protected String nameFormat;
	protected String vlanInfix;
	protected String lastNumber;
	protected String networkName;
	protected String overrideNetmask;
	protected String configDirectory;
	protected String stateDirectory;

	public void load(String key) throws IOException {
		if (configDirectory == null)
			throw new IOException("Directory name not set");

		File f = new File(configDirectory, key + CONFIGFILE_SUFFIX);
		Properties properties = new Properties();
		if (f.exists()) {
			properties.load(new FileReader(f));
		} else {
			throw new IOException("Cannot load " + key + ":" + f.getAbsolutePath() + " not found");
		}

		name = key;
		vlanInfix = properties.getProperty("vlan");
		nameFormat = properties.getProperty("nameFormat");
		overrideNetmask = properties.getProperty("overrideNetmask");
	}

	public List<DeviceType> getAll() throws IOException {
		ArrayList<DeviceType> result = new ArrayList<DeviceType>();

		File d = new File(configDirectory);

		for (File f : d.listFiles()) {
			if (f.isFile() && f.getName().endsWith(CONFIGFILE_SUFFIX)) {
				String name = f.getName();
				int index = name.length() - CONFIGFILE_SUFFIX.length();
				name = name.substring(0, index);
				DeviceType newDeviceType = new DeviceType();
				newDeviceType.setConfigDirectory(configDirectory);
				newDeviceType.setStateDirectory(stateDirectory);
				newDeviceType.load(name);
				result.add(newDeviceType);
			}
		}

		return result;
	}

	public String getVlanAddress(Network network) {
		byte[] addr = Network.parseIpAddress(network.get("ipNetworkNumber"));

		byte n = (byte) Integer.parseInt(vlanInfix);
		addr[2] = n;

		return Network.ipAddressToString(addr);
	}

	public String getNextVlanAddress(Network network) throws NumberFormatException {
		// Network network = new Network();
		// network = (Network) network.fetch(networkName);
		String vlanAddress = getVlanAddress(network);

		byte[] maskBytes;
		byte[] netAddrBytes;
		byte[] addressBytes = Network.parseIpAddress(network.getIpStart());
		byte[] broadcastBytes = new byte[4];

		String ipExcludeRegexp = "dummy";
		Pattern regex = null;

		if (ipExcludeRegexp != null) {
			regex = Pattern.compile(ipExcludeRegexp);
			log.info(ipExcludeRegexp);
		}

		synchronized (this) {
			Network.printIpAddressBytes(addressBytes);
			List<DassComputer> comps = DassComputer.getAllInNetwork(network);

			netAddrBytes = Network.parseIpAddress(vlanAddress);
			
			if(overrideNetmask != null)
				maskBytes = Network.parseIpAddress(overrideNetmask);
			else
				maskBytes = Network.parseIpAddress(network.get("ipNetmaskNumber"));
			
			for (int i = 0; i < 4; i++) {
				broadcastBytes[i] = (byte) (netAddrBytes[i] | (-1 & ~maskBytes[i]));

				addressBytes[i] = (byte) (netAddrBytes[i] | (addressBytes[i] & ~maskBytes[i]));
			}

			Network.printIpAddressBytes(netAddrBytes);
			Network.printIpAddressBytes(maskBytes);
			Network.printIpAddressBytes(broadcastBytes);

			do {
				String addr = Network.ipAddressToString(addressBytes);
				Iterator<DassComputer> iter = comps.iterator();
				boolean collision = false;

				if (regex != null && regex.matcher(addr).matches()) {
					collision = true;
					log.info(addr + " matches exclusion regexp, ignored");
				}
				while (!collision && iter.hasNext()) {
					DassComputer c = iter.next();
					if (c.get("ipHostNumber").equals(addr)) {
						collision = true;
					}
				}

				if (!collision) {
					if (IpDevice.searchByIpAddress(addr) != null)
						collision = true;
				}

				if (!collision) {
					if (DassComputer.searchByIpAddress(addr) != null)
						collision = true;
				}

				if (!collision)
					return addr;

				Network.incAddressBytes(addressBytes);

			} while (!Arrays.equals(addressBytes, broadcastBytes));
		}
		return "127.0.0.1";
	}

	public String getNextName() throws NamingException {
		int nextNumber = 1;

		File nextNumberFile = new File(stateDirectory, "last-" + name);
		BufferedReader reader;
		try {
			reader = new BufferedReader(new FileReader(nextNumberFile));
			nextNumber = Integer.parseInt(reader.readLine());
			reader.close();
		} catch (FileNotFoundException e1) {
			// the file may not exist, ignore
		} catch (NumberFormatException e) {
			log.warning("not a valid integer in " + nextNumberFile.getAbsolutePath());
		} catch (IOException e) {
			log.warning("error reading " + nextNumberFile.getAbsolutePath() + ": " + e.getMessage());
			e.printStackTrace();
		}

		String nextName = String.format(nameFormat, nextNumber);

		log.info("next name " + nextName);
		IpDevice factory = new IpDevice();

		synchronized (DeviceType.class) {
			while (factory.search(IpDevice.getDefaultNamingAttribute(IpDevice.class.getName()) + "=" + nextName,
					IpDevice.getDefaultContainer(IpDevice.class.getName())).size() != 0) {
				nextNumber++;
				nextName = String.format(nameFormat, nextNumber);
				log.info("trying name " + nextName);
			}
		}
		log.info(nextName + " ok");
		FileWriter writer;
		try {
			writer = new FileWriter(nextNumberFile);
			writer.write(nextNumber + "\n");
			writer.close();
		} catch (IOException e) {
			log.warning("error writing " + nextNumberFile.getAbsolutePath() + ": " + e.getMessage());
		}
		return nextName;
	}

	public static void main(String args[]) throws IOException {
		DeviceType newDeviceType = new DeviceType();
		newDeviceType.setConfigDirectory("/tmp/devicetypes");
		newDeviceType.load("test1");

		System.out.println(newDeviceType.getNameFormat());
		System.out.println(newDeviceType.getVlanInfix());

		List<DeviceType> all = newDeviceType.getAll();

		System.out.println("DeviceType getAll:");
		for (DeviceType t : all) {
			System.out.println(t.getName());
		}
		System.out.println("----");

		Network testnet = new Network();
		testnet.setName("testnet");
		testnet.setIpStart("10.33.12.0");

		DeviceType devtype = new DeviceType();
		devtype.setConfigDirectory("/tmp/devicetypes");
		devtype.load("test1");
		System.out.println(devtype.getVlanAddress(testnet));
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getNameFormat() {
		return nameFormat;
	}

	public void setNameFormat(String fmt) {
		this.nameFormat = fmt;
	}

	public String getVlanInfix() {
		return vlanInfix;
	}

	public void setVlanInfix(String vlanInfix) {
		this.vlanInfix = vlanInfix;
	}

	public String getNetworkName() {
		return networkName;
	}

	public void setNetworkName(String networkName) {
		this.networkName = networkName;
	}

	public String getConfigDirectory() {
		return configDirectory;
	}

	public void setConfigDirectory(String configDirectory) {
		this.configDirectory = configDirectory;
	}

	/**
	 * @return the stateDirectory
	 */
	public String getStateDirectory() {
		return stateDirectory;
	}

	/**
	 * @param stateDirectory
	 *            the stateDirectory to set
	 */
	public void setStateDirectory(String stateDirectory) {
		this.stateDirectory = stateDirectory;
	}
}
